from __future__ import annotations

import dataclasses
from typing import Any, Literal, Optional, Union

from .tensors import Value


SolveStatus = Literal[
    "CANCELLED",
    "ERRORED",
    "FEASIBLE",
    "INFEASIBLE",
    "OPTIMAL",
    "UNBOUNDED",
]


@dataclasses.dataclass(frozen=True)
class CancelledOutcome:
    """The solve was cancelled before a solution was found"""


@dataclasses.dataclass(frozen=True)
class FailedOutcome:
    """The solve failed"""

    status: str
    """The underlying error's status"""

    message: str
    """The underlying error's message"""

    code: Optional[str]
    """The underlying error's error code"""

    tags: Any
    """Structured data associated with the failure"""


def failed_outcome_from_graphql(data: Any) -> FailedOutcome:
    failure = data["failure"]
    error = failure["error"]
    return FailedOutcome(
        status=failure["status"],
        message=error["message"],
        code=error.get("code"),
        tags=error.get("tags"),
    )


@dataclasses.dataclass(frozen=True)
class FeasibleOutcome:
    """A solution was found"""

    is_optimal: bool
    """Whether this solution was optimal (within gap thresholds)"""

    objective_value: Optional[Value]
    """The solution's objective value"""

    relative_gap: Optional[Value]
    """The solution's relative gap (0.1 is 10%)"""


def feasible_outcome_from_graphql(data: Any) -> FeasibleOutcome:
    return FeasibleOutcome(
        is_optimal=data["isOptimal"],
        objective_value=data.get("objectiveValue"),
        relative_gap=data.get("relativeGap"),
    )


@dataclasses.dataclass(frozen=True)
class InfeasibleOutcome:
    """No feasible solution exists"""


@dataclasses.dataclass(frozen=True)
class UnboundedOutcome:
    """No bounded optimal solution exists"""


@dataclasses.dataclass(frozen=True)
class AbortedOutcome:
    """No feasible solution was found"""


Outcome = Union[
    AbortedOutcome,
    FailedOutcome,
    FeasibleOutcome,
    InfeasibleOutcome,
    UnboundedOutcome,
]


def outcome_status(outcome: Outcome) -> SolveStatus:
    """Returns the status corresponding to a given outcome"""
    if isinstance(outcome, CancelledOutcome):
        return "CANCELLED"
    if isinstance(outcome, FailedOutcome):
        return "ERRORED"
    if isinstance(outcome, FeasibleOutcome):
        return "OPTIMAL" if outcome.is_optimal else "FEASIBLE"
    if isinstance(outcome, InfeasibleOutcome):
        return "INFEASIBLE"
    if isinstance(outcome, UnboundedOutcome):
        return "INFEASIBLE"
    raise TypeError(f"Unexpected outcome: {outcome}")


class UnexpectedOutcomeError(Exception):
    """The solve ended with an unexpected outcome"""

    def __init__(self, outcome: Outcome):
        self.outcome = outcome



from __future__ import annotations

import collections
import dataclasses
import math
import numpy as np
import pandas as pd
from typing import Any, cast, Mapping, Optional, Sequence, Union

from ..common import decode_extended_float, Json, json_dict
from .outcomes import (
    AbortedOutcome,
    FeasibleOutcome,
    InfeasibleOutcome,
    Outcome,
    SolveStatus,
    UnboundedOutcome,
)
from .outlines import Label, ObjectiveSense, Outline


@dataclasses.dataclass(frozen=True)
class SolveSummary:
    """Reified problem summary statistics"""

    column_count: int
    """Total number of variable columns"""

    row_count: int
    """Total number of constraint rows"""

    dimensions: pd.DataFrame = dataclasses.field(repr=False)
    """Dimension summary statistics"""

    parameters: pd.DataFrame = dataclasses.field(repr=False)
    """Parameter summary statistics"""

    variables: pd.DataFrame = dataclasses.field(repr=False)
    """Variable summary statistics"""

    constraints: pd.DataFrame = dataclasses.field(repr=False)
    """Constraint summary statistics"""

    objectives: pd.DataFrame = dataclasses.field(repr=False)
    """Objective summary statistics"""


def solve_summary_from_json(data: Json) -> SolveSummary:
    return SolveSummary(
        column_count=sum(v["columnCount"] for v in data["variables"]),
        row_count=sum(c["rowCount"] for c in data["constraints"]),
        dimensions=_labeled_dataframe(
            (
                {"label": c["label"], "item_count": c["itemCount"]}
                for c in data["dimensions"]
            )
        ),
        parameters=_labeled_dataframe(
            (
                {
                    "label": c["label"],
                    **_value_profile("entry", c["entryProfile"]),
                    f"entry_{_MULTIPLICITY_SUFFIX}": c["domainMultiplicity"],
                }
                for c in data["parameters"]
            ),
            multiplicities=["entry"],
        ),
        variables=_labeled_dataframe(
            (
                {
                    "label": c["label"],
                    "column_count": c["columnCount"],
                    f"column_{_MULTIPLICITY_SUFFIX}": c["domainMultiplicity"],
                }
                for c in data["variables"]
            ),
            multiplicities=["column"],
        ),
        constraints=_labeled_dataframe(
            (
                {
                    "label": c["label"],
                    "row_count": c["rowCount"],
                    f"row_{_MULTIPLICITY_SUFFIX}": c["domainMultiplicity"],
                    f"row_{_SPARSITY}": math.nan,
                    "column_count": c["columnCount"],
                    f"column_{_MULTIPLICITY_SUFFIX}": c[
                        "coefficientMultiplicity"
                    ],
                    f"column_{_SPARSITY}": math.nan,
                    **_value_profile("weight", c["weightProfile"]),
                    f"weight_{_MULTIPLICITY_SUFFIX}": int(
                        c["domainMultiplicity"]
                    )
                    * int(c["coefficientMultiplicity"]),
                    f"weight_{_SPARSITY}": math.nan,
                    "reify_ms": _timedelta(c["reifiedInMillis"]),
                }
                for c in data["constraints"]
            ),
            multiplicities=["row", "column", "weight"],
        ),
        objectives=_labeled_dataframe(
            (
                {
                    "label": c["label"],
                    **_value_profile("weight", c["weightProfile"]),
                    f"weight_{_MULTIPLICITY_SUFFIX}": c[
                        "coefficientMultiplicity"
                    ],
                    f"weight_{_SPARSITY}": math.nan,
                    "reify_ms": _timedelta(c["reifiedInMillis"]),
                }
                for c in data["objectives"]
            ),
            multiplicities=["weight"],
        ),
    )


_MULTIPLICITY_SUFFIX = "mult"


_SPARSITY = "sprs"


def _labeled_dataframe(
    gen: Any, multiplicities: Optional[Sequence[str]] = None
) -> pd.DataFrame:
    df = pd.DataFrame(gen)
    if not len(df):
        return df
    for k in multiplicities or []:
        m = f"{k}_{_MULTIPLICITY_SUFFIX}"
        se = pd.to_numeric(df[m])
        del df[m]
        with np.errstate(divide="ignore"):
            df[f"{k}_{_SPARSITY}"] = -np.log(df[f"{k}_count"] / se) + 0
    return df.set_index("label")


def _timedelta(ms: int) -> int:
    return ms


_value_profile_keys = ["count", "min", "max", "mean", "stddev"]


def _value_profile(prefix: str, profile: Json) -> Mapping[str, float]:
    return {
        f"{prefix}_{k}": profile.get(k, math.nan) for k in _value_profile_keys
    }


def _entry_index(entries, bindings):
    if not bindings:
        return None
    if len(bindings) == 1:
        binding = bindings[0]
        return pd.Index(
            data=[e["key"][0] for e in entries],
            name=binding.qualifier or binding.dimension_label,
        )
    return pd.MultiIndex.from_tuples(
        tuples=[tuple(e["key"]) for e in entries],
        names=[b.qualifier or b.dimension_label for b in bindings],
    )


@dataclasses.dataclass(frozen=True)
class SolveInputs:
    """Solve input data"""

    outline: Outline
    """Target model metadata"""

    raw_parameters: list[Any]
    """All parameters in raw format"""

    raw_dimensions: Optional[list[Any]]
    """All dimensions in raw format"""

    def parameter(self, label: Label) -> pd.Series:
        """Returns the parameter for a given label as a pandas Series"""
        for param in self.raw_parameters:
            if param["label"] == label:
                entries = param["entries"]
                outline = self.outline.parameters[label]
                return pd.Series(
                    data=(e["value"] for e in entries),
                    index=_entry_index(entries, outline.bindings),
                )
        raise Exception(f"Unknown parameter: {label}")

    def dimension(self, label: Label) -> pd.Index:
        """Returns the dimension for a given label as a pandas Index"""
        for dim in self.raw_dimensions or []:
            if dim["label"] == label:
                return pd.Index(dim["items"])
        raise Exception(f"Unknown dimension: {label}")


@dataclasses.dataclass(frozen=True)
class SolveOutputs:
    """Successful solve output data"""

    outline: Outline
    """Solved model metadata"""

    raw_variables: list[Any]
    """All variables in raw format"""

    raw_constraints: list[Any]
    """All constraints in raw format"""

    def variable(self, label: Label) -> pd.DataFrame:
        """Returns variable results for a given label

        The returned dataframe has two columns: `value` and `dual_value`.
        """
        for res in self.raw_variables:
            if res["label"] == label:
                entries = res["entries"]
                bindings = self.outline.variables[label].bindings
                df = pd.DataFrame(
                    data=(
                        {
                            "value": decode_extended_float(e["value"]),
                            "dual_value": e.get("dualValue"),
                        }
                        for e in entries
                    ),
                    index=_entry_index(entries, bindings),
                )
                return df.dropna(axis=1, how="all").fillna(0)
        raise Exception(f"Unknown variable {label}")

    def constraint(self, label: Label) -> pd.DataFrame:
        """Returns constraint results for a given label.

        The returned dataframe has two columsn: `slack` and `dual_value`.
        """
        for res in self.raw_constraints:
            if res["label"] == label:
                entries = res["entries"]
                outline = self.outline.constraints[label]
                df = pd.DataFrame(
                    data=(
                        {
                            "slack": e["value"],
                            "dual_value": e.get("dualValue"),
                        }
                        for e in entries
                    ),
                    index=_entry_index(entries, outline.bindings),
                )
                return df.dropna(axis=1, how="all").fillna(0)
        raise Exception(f"Unknown constraint {label}")


def _outputs_from_json(data, outline) -> SolveOutputs:
    return SolveOutputs(
        outline=outline,
        raw_variables=data["variables"],
        raw_constraints=data["constraints"],
    )


@dataclasses.dataclass(frozen=True)
class SolveResponse:
    """Solver response"""

    status: SolveStatus
    """Status string"""

    summary: SolveSummary
    """Solve summary statistics"""

    outcome: Outcome
    """Solution metadata"""

    outputs: Optional[SolveOutputs] = dataclasses.field(
        default=None, repr=False
    )
    """Solution data, present iff a solution was found"""


def solve_response_from_json(
    outline: Outline,
    response_json: Any,
    summary: Optional[SolveSummary] = None,
) -> SolveResponse:
    outcome_json = response_json["outcome"]
    status = outcome_json["status"]
    if status == "INFEASIBLE":
        outcome = cast(Outcome, InfeasibleOutcome())
    elif status == "UNBOUNDED":
        outcome = UnboundedOutcome()
    elif status == "ABORTED":
        outcome = AbortedOutcome()
    else:
        outcome = FeasibleOutcome(
            is_optimal=status == "OPTIMAL",
            objective_value=outcome_json.get("objectiveValue"),
            relative_gap=outcome_json.get("relativeGap"),
        )
    outputs = None
    if isinstance(outcome, FeasibleOutcome):
        outputs = _outputs_from_json(
            data=response_json["outputs"],
            outline=outline,
        )
    return SolveResponse(
        status=status,
        outcome=outcome,
        summary=summary or solve_summary_from_json(response_json["summary"]),
        outputs=outputs,
    )


@dataclasses.dataclass(frozen=True)
class SolveOptions:
    """Solving options"""

    relative_gap_threshold: Optional[float] = None
    """Relative gap threshold below which a solution is considered optimal

    For example a value of 0.1 will cause a solution to be optimal when the
    optimality gap is at most 10%. See also `absolute_gap_threshold` for a
    non-relative variant.
    """

    absolute_gap_threshold: Optional[float] = None
    """Absolute gap threshold below which a solution is considered optimal

    See also `relative_gap_threshold` for a relative variant.
    """

    zero_value_threshold: Optional[float] = None
    """Positive magnitude below which tensor values are assumed equal to zero

    This option is also used on solution results, causing values to be omitted
    from the solution if their dual value is also absent. It is finally used as
    threshold for rounding integral variables to the nearest integer. The
    default is 1e-6.
    """

    infinity_value_threshold: Optional[float] = None
    """Positive magnitude used to cap all input values

    It is illegal for the reified problem to include coefficients higher or
    equal to this value so the input needs to be such that they are masked out
    during reification. The default is 1e13.
    """

    free_bound_threshold: Optional[float] = None
    """Positive magnitude used to decide whether a bound is free

    This value should typically be slightly smaller to the infinity value
    threshold to allow for small offsets to infinite values. The default is
    1e12.
    """

    timeout_millis: Optional[float] = None
    """Upper bound on solving time"""


def solve_options_to_json(options: Optional[SolveOptions] = None) -> Json:
    if not options:
        return None
    return json_dict(**dataclasses.asdict(options or SolveOptions()))


Target = Union[Label, Mapping[Label, float]]
"""Target objective

A single label is equivalent to optimizing just the objective with that label
and ignoring all others. If using a mapping, all objective keys must have an
associated values.
"""


def _target_to_json(target: Target, outline: Outline) -> Json:
    if isinstance(target, str):
        target = collections.defaultdict(lambda: 0, {target: 1})
    unknown = target.keys() - outline.objectives.keys()
    if unknown:
        raise Exception(f"Unknown objective(s): {unknown}")
    weights = [
        {"label": label, "value": target[label]}
        for label in outline.objectives
    ]
    return json_dict(weights=weights)


@dataclasses.dataclass(frozen=True)
class EpsilonConstraint:
    """Constraint enforcing proximity to a objective's optimal value"""

    target: Target
    """Target objective"""

    absolute_tolerance: Optional[float] = None
    """Cap on the absolute value of the final solution vs optimal"""

    relative_tolerance: Optional[float] = None
    """Cap on the relative value of the final solution vs optimal"""


@dataclasses.dataclass(frozen=True)
class SolveStrategy:
    """Multi-objective solving strategy"""

    target: Target
    """Target objective"""

    sense: Optional[ObjectiveSense] = None
    """Optimization sense"""

    epsilon_constraints: list[EpsilonConstraint] = dataclasses.field(
        default_factory=lambda: []
    )
    """All epsilon-constraints to apply"""

    @classmethod
    def equally_weighted_sum(cls, sense: Optional[ObjectiveSense] = None):
        """Returns a strategy optimizing the sum of all objectives"""
        return SolveStrategy(
            target=collections.defaultdict(lambda: 1), sense=sense
        )


def solve_strategy_to_json(
    strategy: Optional[SolveStrategy], outline: Outline
) -> Json:
    if not strategy:
        return None
    target = strategy.target
    if isinstance(target, str):
        target = collections.defaultdict(lambda: 0, {target: 1})
    sense = strategy.sense
    if not sense:
        for label, objective in outline.objectives.items():
            weight = target[label]
            if not weight:
                continue
            if sense is None:
                sense = objective.sense
            elif sense != objective.sense:
                raise Exception("Explicit objective sense required")
        if not sense:
            raise Exception("Missing objective")
    return json_dict(
        is_maximization=sense == "MAXIMIZE",
        target=_target_to_json(target, outline),
        epsilon_constraints=[
            json_dict(
                relative_tolerance=c.relative_tolerance,
                absolute_tolerance=c.absolute_tolerance,
                target=_target_to_json(c.target, outline),
            )
            for c in strategy.epsilon_constraints
        ],
    )



from __future__ import annotations

import dataclasses
import math
import pandas as pd
from typing import Any, Iterable, Mapping, Tuple, Union

from ..common import ExtendedFloat, encode_extended_float


KeyItem = Union[float, int, str]


Key = Union[Tuple[KeyItem, ...], KeyItem]


Value = Union[float, int]


def is_value(arg: Any) -> bool:
    return isinstance(arg, (float, int))


DimensionArgument = Iterable[KeyItem]


SparseTensorArgument = Union[
    Mapping[Key, Value],
    pd.Series,
    pd.DataFrame,  # For indicator parameters
    Iterable[Key],  # For indicator parameters
]


#: Input tensor creation argument
TensorArgument = Union[
    Value, SparseTensorArgument, Tuple[SparseTensorArgument, Value]
]


@dataclasses.dataclass
class Tensor:
    """An n-dimensional matrix"""

    entries: list[Any]
    """Raw list of matrix entries"""

    default_value: ExtendedFloat = 0
    """Value to use for missing key"""

    @classmethod
    def from_argument(
        cls,
        arg: TensorArgument,
        rank: int,
        is_indicator: bool = False,
        is_pin: bool = False,
    ):
        """Creates a tensor from a variety of argument values

        In most cases you will not need to call this method directly: it is
        called automatically during parameter validation.

        Args:
            arg: Raw argument to be wrapped into a tensor. This argument must
                match the tensor's shape, see below for details.
            rank: The expected rank of the tensor
            is_indicator: Whether the tensor holds indicator values

        The accepted arguments depend on the tensor's domain rank (the length
        of its keys) and image (does it hold arbitrary numbers or only 0s/1s).
        If the tensor's rank is 0, i.e. is a scalar, the only accepted argument
        type is a number matching the image. Otherwise the following inputs are
        accepted:

        + Mapping (e.g. Python dictionary) with tuple key if rank is greater
          than 1
        + `pandas` series with key index

        Additionally, if the tensor holds indicator values, two more
        conveniences are allowed, each representing the set of keys which have
        value 1:

        + Iterable of keys
        + `pandas` dataframe

        Finally, all non-scalar tensor arguments can be wrapped into a tuple
        `(arg, default)` to provide a `default` value to use when no matching
        key exists.
        """
        if isinstance(arg, tuple):
            data, default_value = arg
        elif rank > 0 and is_value(arg) and not is_indicator:
            data = {}
            default_value = arg
        else:
            data = arg
            default_value = -math.inf if is_pin else 0
        if (
            is_indicator
            and isinstance(data, pd.Series)
            and not pd.api.types.is_numeric_dtype(data)
        ):
            data = data.reset_index()
        keyifier = _Keyifier(rank)
        if is_indicator and isinstance(data, pd.DataFrame):
            entries = [
                {"key": keyifier.keyify(key), "value": 1}
                for key in data.itertuples(index=False, name=None)
            ]
        elif is_indicator and not hasattr(data, "items"):
            entries = [
                {"key": keyifier.keyify(key), "value": 1} for key in data
            ]
        else:
            if is_value(data):
                entries = [{"key": (), "value": encode_extended_float(data)}]
            else:
                entries = [
                    {
                        "key": keyifier.keyify(key),
                        "value": encode_extended_float(value),
                    }
                    for key, value in data.items()
                ]
        return Tensor(entries, encode_extended_float(default_value))


class _Keyifier:
    def __init__(self, rank: int):
        self.rank = rank

    def keyify(self, key):
        tup = tuple(key) if isinstance(key, (list, tuple)) else (key,)
        if len(tup) != self.rank:
            raise Exception(f"Invalid key rank: {len(tup)} != {self.rank}")
        return tup



from __future__ import annotations

import dataclasses
import functools
import inspect
import itertools
import logging
import math
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Literal,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from ..common import Label
from .ast import (
    Domain,
    Expression,
    ExpressionLike,
    ExpressionReference,
    IterableSpace,
    Predicate,
    Quantifiable,
    QuantifiableReference,
    Quantifier,
    QuantifierIdentifier,
    ScalarSpace,
    Space,
    cross,
    domain,
    expression_space,
    is_literal,
    literal,
    render_identifier,
    to_expression,
    within_domain,
)
from .identifiers import (
    AliasIdentifier,
    DimensionIdentifier,
    GlobalIdentifier,
    Name,
    TensorIdentifier,
    local_formatting_scope,
)
from .quantified import Quantified, unquantify
from .statements import Definition, Model, ModelFragment, method_decorator


_logger = logging.getLogger(__name__)


class Dimension(Definition, ScalarSpace):
    """An abstract collection of values

    Args:
        label: Dimension label override. By default the label is derived from
            the attribute's name
        name: The dimension's name. By default the name will be derived from
            the dimension's label
        is_numeric: Whether the dimension will only contain integers. This
            enables arithmetic operations on this dimension's quantifiers

    Dimensions are `Quantifiable` and as such can be quantified over using
    :func:`.cross`. As a convenience, iterating on a dimension also yields a
    suitable quantifier. This allows creating simple constraints directly:

    .. code-block:: python

        class ProductModel(Model):
            products = Dimension()
            count = Variable.natural(products)

            @constraint
            def at_least_one_of_each(self):
                for p in self.products:  # Note the iteration here
                    yield self.count(p) >= 1
    """

    category = "DIMENSION"

    def __init__(
        self,
        *,
        label: Optional[Label] = None,
        name: Optional[Name] = None,
        is_numeric: bool = False,
    ):
        self._identifier = DimensionIdentifier(name=name)
        self._label = label
        self._is_numeric = is_numeric

    @property
    def identifier(self) -> Optional[GlobalIdentifier]:
        return self._identifier

    @property
    def label(self) -> Optional[Label]:
        return self._label

    def render(self) -> str:
        return self._identifier.format()

    def render_statement(self, label: Label) -> Optional[str]:
        _logger.debug("Rendering dimension %s...", label)
        s = f"\\S^d_\\mathrm{{{label}}}&: {self._identifier.format()}"
        if self._is_numeric:
            s += " \\subset \\mathbb{Z}"
        return s


@dataclasses.dataclass(frozen=True)
class _Interval(ScalarSpace):
    lower_bound: Expression
    upper_bound: Expression

    def render(self) -> str:
        lb = self.lower_bound
        ub = self.upper_bound
        if is_literal(lb, 0) and is_literal(ub, 1):
            return "\\{0, 1\\}"
        if is_literal(ub, math.inf):
            if is_literal(lb, 0):
                return "\\mathbb{N}"
            elif is_literal(lb, -math.inf):
                return "\\mathbb{Z}"
        return f"\\{{ {lb.render()} \\ldots {ub.render()} \\}}"


_integers = _Interval(literal(-math.inf), literal(math.inf))


def interval(
    lower_bound: ExpressionLike,
    upper_bound: ExpressionLike,
    name: Optional[Name] = None,
) -> IterableSpace[Quantifier]:
    """A range of values

    Args:
        lower_bound: The range's inclusive lower bound
        upper_bound: The range's inclusive upper bound
        name: An optional name to create an alias for this interval
    """
    interval = _Interval(
        lower_bound=to_expression(lower_bound),
        upper_bound=to_expression(upper_bound),
    )

    class _Fragment(ModelFragment, Space):
        @property
        @alias(name)
        def interval(self):
            return interval

        def __iter__(self) -> Iterator[Quantifier]:
            return iter(self.interval)

    return _Fragment()


@dataclasses.dataclass(frozen=True)
class Image:
    """A tensor's set of possible values"""

    lower_bound: ExpressionLike = -math.inf
    """The image's smallest value (inclusive)"""

    upper_bound: ExpressionLike = math.inf
    """The image's largest value (inclusive)"""

    is_integral: bool = False
    """Whether the image only contain integers"""

    def render(self) -> str:
        lb = to_expression(self.lower_bound)
        ub = to_expression(self.upper_bound)
        if is_literal(ub, math.inf):
            if is_literal(lb, -math.inf):
                return "\\mathbb{Z}" if self.is_integral else "\\mathbb{R}"
            if is_literal(lb, 0):
                return "\\mathbb{N}" if self.is_integral else "\\mathbb{R}_+"
        if self.is_integral:
            if is_literal(lb, 0) and is_literal(ub, 1):
                return "\\{0, 1\\}"
            return f"\\{{{lb.render()} \\ldots {ub.render()}\\}}"
        return f"[{lb.render()}, {ub.render()}]"


TensorLike = Callable[..., Expression]


_T = TypeVar("_T", bound="Tensor")


class Tensor(Definition):
    """Base tensor class

    Calling a tensor returns an :class:`~.opvious.modeling.Expression` with any
    arguments as subscripts. For example:

    .. code-block:: python

        class ProductModel(Model):
            products = Dimension()
            count = Variable.natural(products)

            @property
            def total_product_count(self):
                return total(
                    self.count(p)  # Expression representing the count for `p`
                    for p in self.products
                )

    The number of arguments must match the tensor's quantification.
    """

    def __init__(
        self,
        image: Image,
        *quantifiables: Quantifiable,
        name: Optional[Name] = None,
        label: Optional[Label] = None,
        qualifiers: Optional[Sequence[Label]] = None,
    ):
        if not isinstance(image, Image):
            raise TypeError(f"Unexpected image: {image}")
        self._identifier = TensorIdentifier(
            name=name,
            is_parameter=self.category == "PARAMETER",
        )
        # We do not combine domains here to allow quantification projections in
        # an alias-agnostic (and simpler) way
        self._domains = tuple(domain(q) for q in quantifiables)
        self._label = label
        self._image = image
        self.qualifiers = qualifiers

    @classmethod
    def continuous(cls: Type[_T], *quantifiables, **kwargs) -> _T:
        """Returns a tensor with real image"""
        return cls(Image(), *quantifiables, **kwargs)

    @classmethod
    def non_negative(
        cls: Type[_T],
        *quantifiables,
        upper_bound: ExpressionLike = math.inf,
        **kwargs,
    ) -> _T:
        """Returns a tensor with non-negative real image"""
        img = Image(lower_bound=0, upper_bound=upper_bound)
        return cls(img, *quantifiables, **kwargs)

    @classmethod
    def non_positive(cls: Type[_T], *quantifiables, **kwargs) -> _T:
        """Returns a tensor with non-positive real image"""
        return cls(Image(upper_bound=0), *quantifiables, **kwargs)

    @classmethod
    def unit(cls: Type[_T], *quantifiables, **kwargs) -> _T:
        """Returns a tensor with `[0, 1]` real image"""
        return cls(
            Image(lower_bound=0, upper_bound=1), *quantifiables, **kwargs
        )

    @classmethod
    def discrete(cls: Type[_T], *quantifiables, **kwargs) -> _T:
        """Returns a tensor with integral image"""
        return cls(Image(is_integral=True), *quantifiables, **kwargs)

    @classmethod
    def natural(
        cls: Type[_T],
        *quantifiables,
        upper_bound: ExpressionLike = math.inf,
        **kwargs,
    ) -> _T:
        """Returns a tensor with non-negative integral image"""
        img = Image(lower_bound=0, upper_bound=upper_bound, is_integral=True)
        return cls(img, *quantifiables, **kwargs)

    @classmethod
    def indicator(cls: Type[_T], *quantifiables, **kwargs) -> _T:
        """Returns a tensor with `{0, 1}` integral image"""
        image = Image(lower_bound=0, upper_bound=1, is_integral=True)
        return cls(image, *quantifiables, **kwargs)

    @property
    def identifier(self) -> Optional[GlobalIdentifier]:
        return self._identifier

    @property
    def label(self) -> Optional[Label]:
        return self._label

    @property
    def image(self) -> Image:
        """The tensor's values' image"""
        return self._image

    def quantifiables(self) -> tuple[Quantifiable, ...]:
        """The quantifiables generating the tensor's space

        This is typically useful to support projections in model fragments. For
        most use cases consider using the simpler
        :meth:`~opvious.modeling.Tensor.space` method.
        """
        return self._domains

    def space(self) -> IterableSpace[Sequence[Quantifier]]:
        """Returns the tensor's underlying space"""
        return cross(self._domains)

    def __call__(self, *subscripts: ExpressionLike) -> Expression:
        return ExpressionReference(
            self._identifier, tuple(to_expression(s) for s in subscripts)
        )

    def render_statement(self, label: Label) -> Optional[str]:
        _logger.debug("Rendering tensor %s...", label)
        c = self.category[0].lower()
        s = f"\\S^{c}_\\mathrm{{{_render_label(label, self.qualifiers)}}}&: "
        s += f"{self._identifier.format()} \\in "
        s += self.image.render()
        d = domain(self._domains)
        if d.quantifiers:
            with local_formatting_scope(d.quantifiers):
                if _is_simple_domain(d):
                    formatted: list[str] = []
                    for g, qs in itertools.groupby(
                        d.quantifiers, key=lambda q: q.outer_group
                    ):
                        if g is None:
                            formatted.extend(q.space.render() for q in qs)
                        else:
                            formatted.append(g.alias.format())
                    sup = " \\times ".join(formatted)
                else:
                    sup = f"\\{{ {d.render()} \\}}"
                s += f"^{{{sup}}}"
        return s


def _is_simple_domain(d: Domain) -> bool:
    if d.mask is not None:
        return False
    for q in d.quantifiers:
        for g in q.groups:
            if g.subscripts:
                return False
    return True


def _render_label(label: Label, qualifiers: Optional[Sequence[Label]]) -> str:
    s = label
    if qualifiers:
        s += f"[{','.join(qualifiers)}]"
    return s


class Parameter(Tensor):
    """An optimization input parameter

    Args:
        image: Target :class:`~opvious.modeling.Image`
        quantifiables: Quantification
        label: Optional label, if unset it will be derived from the parameter's
            attribute name
        name: Optional name, if unset it will be derived from the labe
        qualifiers: Optional list of labels used to name this parameter's
            quantifiables

    Consider instantiating parameters via one of the various
    :class:`~opvious.modeling.Tensor` convenience class methods, for example:

    .. code-block:: python

        p1 = Parameter.continuous()  # Real-valued parameter
        p2 = Parameter.natural() # # Parameter with values in {0, 1...}

    Each parameter will be expected to have a matching input when solving this
    model on data.
    """

    category = "PARAMETER"


class Variable(Tensor):
    """An optimization output variable

    Args:
        image: Target :class:`~opvious.modeling.Image`
        quantifiables: Quantification
        label: Optional label, if unset it will be derived from the parameter's
            attribute name
        name: Optional name, if unset it will be derived from the labe
        qualifiers: Optional list of labels used to name this parameter's
            quantifiables

    Similar to parameters, consider instantiating variables via one of the
    various :class:`~opvious.modeling.Tensor` convenience class methods, for
    example:

    .. code-block:: python

        v1 = Variable.unit()  # Variable with value within [0, 1]
        v2 = Variable.non_negative() # # Variable with value at least 0

    Each variable will have its results available in feasible solve solutions.
    """

    category = "VARIABLE"


_Aliasable = Callable[..., Any]


@dataclasses.dataclass(frozen=True)
class _Aliased:
    quantifiable: Quantifiable
    quantifiers: Union[
        None, QuantifierIdentifier, tuple[QuantifierIdentifier, ...]
    ]


class _Alias(Definition):
    category = "ALIAS"
    label = None

    def __init__(
        self,
        aliasable: _Aliasable,
        quantifiable: tuple[Quantifiable],
        name: Name,
        quantifier_names: Optional[Iterable[Name]] = None,
    ):
        super().__init__()
        self._identifier = AliasIdentifier(name=name)
        self._quantifier_names = quantifier_names
        self._aliasable = aliasable
        self._aliased: Optional[_Aliased] = None
        self._quantifiable = quantifiable

    @property
    def identifier(self) -> Optional[GlobalIdentifier]:
        return self._identifier

    def render_statement(self, _label: Label) -> Optional[str]:
        if self._aliased is None:
            _logger.debug("Skipping alias named %s.", self._identifier.name)
            return None  # Not used

        _logger.debug("Rendering alias named %s...", self._identifier.name)
        outer_domain = domain(
            self._aliased.quantifiable, names=self._quantifier_names
        )
        expressions = [Quantifier(q) for q in outer_domain.quantifiers]
        value = self._aliasable(*expressions)

        s = "\\S^a&: "
        with local_formatting_scope(outer_domain.quantifiers):
            if outer_domain.quantifiers:
                s += f"\\forall {outer_domain.render()}, "
            s += render_identifier(self._identifier, *expressions)
            s += " \\doteq "
            if self._aliased.quantifiers is None:
                s += value.render()
            else:
                inner_domain = domain(value)
                with local_formatting_scope(inner_domain.quantifiers):
                    if (
                        len(inner_domain.quantifiers) > 1
                        or inner_domain.mask is not None
                    ):
                        s += f"\\{{ {inner_domain.render()} \\}}"
                    else:
                        s += inner_domain.quantifiers[0].space.render()
        return s

    def __call__(self, *subs: ExpressionLike) -> Any:
        exprs = tuple(to_expression(s) for s in subs)
        if self._aliased is None:
            value = self._aliasable(*exprs)
            if isinstance(value, Expression):
                quantifiers = None
            else:
                quantifiers, _ = unquantify(value)
            self._aliased = _Aliased(
                self._quantifiable
                or tuple(expression_space(x) or _integers for x in exprs),
                None
                if quantifiers is None
                else _quantifier_identifier(quantifiers)
                if isinstance(quantifiers, Quantifier)
                else tuple(_quantifier_identifier(q) for q in quantifiers),
            )
        quantifiers = self._aliased.quantifiers
        if quantifiers is None:
            return ExpressionReference(self._identifier, exprs)
        else:
            ref = QuantifiableReference(
                identifier=self._identifier,
                subscripts=exprs,
                quantifiers=quantifiers
                if isinstance(quantifiers, tuple)
                else (quantifiers,),
            )
            if isinstance(self._aliased.quantifiers, tuple):
                return cross(ref)
            else:
                return iter(ref)


def _quantifier_identifier(arg: Any) -> QuantifierIdentifier:
    if not isinstance(arg, Quantifier):
        raise TypeError(f"Space aliases should only return quantifiers: {arg}")
    return arg.identifier


_M = TypeVar("_M", bound=Union[Model, ModelFragment], contravariant=True)


_F = TypeVar("_F", bound=Callable[..., Union[Expression, Quantifiable]])


def alias(
    name: Optional[Name],
    *quantifiables: Quantifiable,
    quantifier_names: Optional[Iterable[Name]] = None,
) -> Callable[[_F], _F]:  # TODO: Tighten argument type
    """Decorator promoting a :class:`.Model` method to a named alias

    Args:
        name: The alias' name. If `None`, no alias will be added.
        quantifiables: The alias' quantification. If omitted, it will be
            inferred from the alias' calls.
        quantifier_names: Optional names to use for the alias' quantifiers

    The method can return a (potentially quantified) expression or a
    quantification and may accept any number of expression arguments. This is
    useful to make the generated specification more readable by extracting
    commonly used sub-expressions or sub-spaces.

    Finally, the decorated function may be wrapped as a property if doesn't
    have any non-`self` arguments.

    .. code-block:: python

        class ProductModel(Model):
            products = Dimension()
            count = Variable.natural(products)

            @property
            @alias("t")
            def total_product_count(self):
                return total(self.count(p) for p in self.products)
    """

    @method_decorator
    def wrapper(fn):
        if name is None:
            return fn
        return _Alias(
            fn,
            quantifiable=quantifiables,
            name=name,
            quantifier_names=quantifier_names,
        )

    return wrapper


ConstraintMethod = Callable[[_M], Quantified[Predicate]]


class Constraint(Definition):
    """Optimization constraint

    Constraints are best created directly from :class:`.Model` methods via the
    :func:`.constraint` decorator.

    .. code-block:: python

        class ProductModel:
            products = Dimension()
            count = Variable.natural(products)

            @constraint
            def at_least_one(self):
                yield total(self.count(p) for p in self.products) >= 1
    """

    category = "CONSTRAINT"
    identifier = None

    def __init__(
        self,
        body: Callable[[], Quantified[Predicate]],
        label: Optional[Label] = None,
        qualifiers: Optional[Sequence[Label]] = None,
    ):
        super().__init__()
        self._body = body
        self._label = label
        self.qualifiers = qualifiers

    @property
    def label(self):
        return self._label

    def render_statement(self, label: Label) -> Optional[str]:
        _logger.debug("Rendering constraint %s...", label)

        s = f"\\S^c_\\mathrm{{{_render_label(label, self.qualifiers)}}}&: "
        predicate, d = within_domain(self._body())
        with local_formatting_scope(d.quantifiers):
            if d.quantifiers:
                s += f"\\forall {d.render()}, "
            s += predicate.render()
        return s


@overload
def constraint(method: ConstraintMethod) -> Constraint:
    ...


@overload
def constraint(
    *,
    label: Optional[Label] = None,
    qualifiers: Optional[Sequence[Label]] = None,
    disabled=False,
) -> Callable[[ConstraintMethod], Optional[Constraint]]:
    ...


def constraint(
    method: Optional[ConstraintMethod] = None,
    *,
    label: Optional[Label] = None,
    qualifiers: Optional[Sequence[Label]] = None,
    disabled=False,
) -> Any:
    """Decorator promoting a :class:`.Model` method to a :class:`.Constraint`

    Args:
        label: Constraint label override. By default the label is derived from
            the method's name.
        qualifiers: Optional list of labels used to qualify the constraint's
            quantifiers. This is useful to override the name of the colums in
            solution dataframes.
        disabled: Disable the constraint

    The decorated method should accept only a `self` argument and return a
    quantified :class:`.Predicate`.

    As a convenience, this decorator can be used with and without arguments:

    .. code-block:: python

        class MyModel(Model):
            # ...

            @constraint
            def ensure_something(self):
                # ...

            @constraint(label="ensuresEverything")
            def ensure_something_else(self):
                # ...
    """

    @method_decorator
    def wrapper(fn):
        if not inspect.isgeneratorfunction(fn):
            raise TypeError(f"Non-generator constraint function: {fn}")
        if disabled:
            return None
        return Constraint(fn, label=label, qualifiers=qualifiers)

    return wrapper(method) if method else wrapper


ObjectiveSense = Literal["max", "min"]
"""Optimization direction"""


ObjectiveMethod = Callable[[_M], Expression]
"""Optimization target expression"""


class Objective(Definition):
    """Optimization objective

    Objectives are best created directly from :class:`.Model` methods via the
    :func:`.objective` decorator.

    .. code-block:: python

        class ProductModel:
            products = Dimension()
            count = Variable.natural(products)

            @objective
            def minimize_total_count(self):
                return total(self.count(p) for p in self.products)

    """

    category = "OBJECTIVE"
    identifier = None

    def __init__(
        self,
        body: Callable[[], Expression],
        sense: ObjectiveSense,
        label: Optional[Label] = None,
    ):
        super().__init__()
        self._body = body
        self._sense = sense
        self._label = label

    @property
    def label(self) -> Optional[Label]:
        return self._label

    def render_statement(self, label: Label) -> Optional[str]:
        _logger.debug("Rendering objective %s...", label)

        sense = self._sense
        if sense is None:
            if label.startswith("min"):
                sense = "min"
            elif label.startswith("max"):
                sense = "max"
            else:
                raise Exception(f"Missing sense for objective {label}")

        expression = to_expression(self._body())
        return f"\\S^o_\\mathrm{{{label}}}&: \\{sense} {expression.render()}"


@overload
def objective(method: ObjectiveMethod) -> Objective:
    ...


@overload
def objective(
    *,
    sense: Optional[ObjectiveSense] = None,
    label: Optional[Label] = None,
    disabled=False,
) -> Callable[[ObjectiveMethod], Objective]:
    ...


def objective(
    method: Optional[ObjectiveMethod] = None,
    *,
    sense: Optional[ObjectiveSense] = None,
    label: Optional[Label] = None,
    disabled=False,
) -> Any:
    """Decorator promoting a method to an :class:`.Objective`

    Args:
        sense: Optimization direction. This may be omitted if the method name
            starts with `min` or `max`, in which case the appropriate sense
            will be inferred.
        label: Objective label override. By default the label is derived from
            the method's name.
        disabled: Disable the objective

    The decorated method should accept only a `self` argument and return an
    :class:`.Expression`, which will become the objective's optimization
    target.

    As a convenience, this decorator can be used with and without arguments:

    .. code-block:: python

        class ProductModel(Model):
            products = Dimension()
            cost = Parameter.non_negative(products)
            count = Variable.natural(products)

            @objective
            def minimize_cost(self):
                return total(
                    self.count(p) * self.cost(p)
                    for p in self.products
                )

            @objective(sense="max")
            def optimize_count(self):
                return total(self.count(p) for p in self.products)
    """

    @method_decorator
    def wrapper(fn):
        if disabled:
            return None
        method_sense = sense or _objective_sense(fn)
        return Objective(fn, sense=method_sense, label=label)

    return wrapper(method) if method else wrapper


def _objective_sense(fn: Callable[..., Any]) -> ObjectiveSense:
    while isinstance(fn, functools.partial):
        fn = fn.func
    name = fn.__name__
    if name.startswith("min"):
        return "min"
    elif name.startswith("max"):
        return "max"
    else:
        raise Exception(f"Missing sense for objective {name}")



from __future__ import annotations

import collections
import dataclasses
import functools
import logging
import pandas as pd
from typing import (
    Any,
    Callable,
    Iterable,
    Literal,
    Mapping,
    Optional,
    Sequence,
)
import weakref

from ..common import Label, to_camel_case
from ..specifications.local import LocalSpecification, LocalSpecificationSource
from .identifiers import (
    DefaultIdentifierFormatter,
    GlobalIdentifier,
    Name,
    global_formatting_scope,
)


_logger = logging.getLogger(__name__)


DefinitionCategory = Literal[
    "ALIAS",
    "CONSTRAINT",
    "DIMENSION",
    "OBJECTIVE",
    "PARAMETER",
    "VARIABLE",
]


class Definition:
    """Base model definition"""

    @property
    def category(self) -> DefinitionCategory:
        raise NotImplementedError()

    @property
    def label(self) -> Optional[Label]:
        raise NotImplementedError()

    @property
    def identifier(self) -> Optional[GlobalIdentifier]:
        raise NotImplementedError()

    def render_statement(self, label: Label) -> Optional[str]:
        raise NotImplementedError()


class ModelFragment:
    """Reusable model sub-component

    Model fragments are useful to group related definitions together and expose
    them in a reusable way. See :ref:`the API reference
    <\\`opvious.modeling.fragments\\`>` for the list of available fragments.
    """

    @property
    def default_definition(self) -> Optional[str]:
        return None


def method_decorator(wrapper: Callable[..., Any]) -> Any:
    def wrap(fn):
        return _DecoratedMethod(fn, wrapper)

    return wrap


class _DecoratedMethod:
    def __init__(
        self,
        body: Callable[..., Any],
        wrapper: Callable[[Callable[..., Any]], Any],
    ) -> None:
        self._body = body
        self._wrapper = wrapper
        self._bindings: Any = weakref.WeakKeyDictionary()

    def bound_to(self, owner: Any) -> Any:
        binding = self._bindings.get(owner)
        if not binding:
            binding = self._wrapper(functools.partial(self._body, owner))
            self._bindings[owner] = binding
        return binding

    def __get__(self, owner: Any, _objtype=None) -> Any:
        return self.bound_to(owner)

    def __call__(self, owner) -> Any:  # Property call
        return self.bound_to(owner)()


@dataclasses.dataclass(frozen=True)
class Statement:
    """A rendered definition"""

    title: str
    """The title of the model the definition belongs to"""

    category: DefinitionCategory
    """The type of definition this statement contains"""

    label: Label
    """The definition's label"""

    name: Optional[Name]
    """The definition's name, if applicable"""

    text: str
    """The definition's mathematical representation"""


@dataclasses.dataclass(frozen=True)
class _Candidate:
    label: Label
    definition: Definition
    model: Model

    def render_statement(self) -> Optional[Statement]:
        d = self.definition
        text = d.render_statement(self.label)
        if text is None:
            return None
        return Statement(
            title=self.model.title,
            category=d.category,
            label=self.label,
            name=d.identifier.format() if d.identifier else None,
            text=text,
        )


@dataclasses.dataclass(frozen=True)
class _Relabeled(ModelFragment):
    fragment: ModelFragment
    labels: Mapping[str, Label]


def relabel(fragment: ModelFragment, **kwargs: Label) -> ModelFragment:
    """Updates a fragment's definitions' labels

    Args:
        fragment: The fragment containing definitions to relabel
        kwargs: Dictionary from old label to new label
    """
    return _Relabeled(fragment, kwargs)


class _ModelVisitor:
    def __init__(self) -> None:
        self.candidates: list[_Candidate] = []
        self._visited: set[int] = set()  # Fragment IDs

    def visit(self, model: Model) -> None:
        model_id = id(model)
        if model_id in self._visited:
            return
        self._visited.add(model_id)
        self._visit_owner(model=model, fragment=None, prefix=model.prefix)
        for dep in model.dependencies:
            self.visit(dep)

    def _visit_owner(
        self,
        model: Model,
        fragment: Optional[ModelFragment],
        prefix: Sequence[str],
    ) -> None:
        owner = fragment or model

        labels: dict[str, Label] = {}
        while isinstance(owner, _Relabeled):
            labels.update(owner.labels)
            owner = owner.fragment

        attrs: dict[str, Any] = {}
        for cls in reversed(owner.__class__.__mro__[1:]):
            attrs.update(cls.__dict__)
        attrs.update(owner.__dict__)
        attrs.update(owner.__class__.__dict__)

        path = [*prefix, ""]
        for attr, value in attrs.items():
            if attr.startswith("_"):
                continue
            if fragment and attr == fragment.default_definition:
                path[-1] = ""
            else:
                path[-1] = attr
            if isinstance(value, property):
                value = value.fget
            if isinstance(value, _DecoratedMethod):
                value = value.bound_to(owner)
            if isinstance(value, ModelFragment):
                self._visit_owner(model, value, path)
                continue
            if not isinstance(value, Definition):
                continue
            label = (
                labels.get(attr)
                or value.label
                or to_camel_case("_".join(path))
            )
            self.candidates.append(_Candidate(label, value, model))


class Model:
    """An optimization model

    Args:
        dependencies: Optional list of models upon which this model's
            definitions depend. Dependencies' definitions will be automatically
            added when generating this model's specification.
        prefix: Optional prefix added to all generated labels in this model
        title: Optional title used when creating the model's
            :class:`.LocalSpecification`
    """

    __dependencies: Optional[Sequence[Model]] = None
    __prefix: Optional[Sequence[str]] = None
    __title: Optional[str] = None

    def __init__(
        self,
        dependencies: Optional[Iterable[Model]] = None,
        prefix: Optional[Sequence[str]] = None,
        title: Optional[str] = None,
    ):
        self.__dependencies = list(dependencies) if dependencies else None
        self.__prefix = prefix
        self.__title = title

    @property
    def dependencies(self) -> Sequence[Model]:
        """The model's list of dependencies"""
        return self.__dependencies or []

    @property
    def prefix(self) -> Sequence[str]:
        """The model's label prefix, or an empty list if unset"""
        return self.__prefix or []

    @property
    def title(self) -> str:
        """The model's title, defaulting to its class' `__qualname__`"""
        return self.__title or self.__class__.__qualname__

    def statements(self) -> Iterable[Statement]:
        """Lists the model's (and any dependencies') statements"""
        visitor = _ModelVisitor()
        visitor.visit(self)
        candidates = visitor.candidates

        by_identifier = {
            c.definition.identifier: c
            for c in candidates
            if c.definition.identifier
        }
        labels_by_identifier = {
            i: d.label for i, d in by_identifier.items() if d.label
        }
        formatter = DefaultIdentifierFormatter(labels_by_identifier)
        reserved = {i.name: i for i in labels_by_identifier if i.name}
        with global_formatting_scope(formatter, reserved):
            idens = set()
            for c in candidates:
                s = c.render_statement()
                if not s:
                    continue
                yield s
                if c.definition.identifier:
                    idens.add(c.definition.identifier)
            for iden in formatter.formatted_globals():
                if iden in idens:
                    continue
                dc = by_identifier.get(iden)
                if not dc:
                    raise Exception(f"Missing candidate: {iden}")
                s = dc.render_statement()
                if not s:
                    raise Exception(f"Missing statement: {iden}")
                yield s

    def definition_counts(self) -> pd.DataFrame:
        """Returns a dataframe summarizing the number of definitions"""
        df = pd.DataFrame(dataclasses.asdict(s) for s in self.statements())
        grouped: Any = df.groupby(["title", "category"])["text"].count()
        return grouped.unstack(["category"]).fillna(0).astype(int)

    def specification(self) -> LocalSpecification:
        """Generates the model's specification

        This specification can be used to interact with :class:`.Client`
        methods, for example to start a solve.
        """
        grouped = collections.defaultdict(list)
        for s in self.statements():
            grouped[s.title].append(s.text)
        joined = {
            title: "".join(f"  {s} \\\\\n" for s in lines)
            for title, lines in grouped.items()
        }
        sources = [
            LocalSpecificationSource(
                title=title,
                text=f"$$\n\\begin{{align*}}\n{contents}\\end{{align*}}\n$$",
            )
            for title, contents in joined.items()
        ]
        return LocalSpecification(sources=sources, description=self.__doc__)



import base64
import sys
from typing import Optional

from .common import (
    Executor,
    ExecutorError,
    ExecutorResult,
    JsonExecutorResult,
    JsonSeqExecutorResult,
    PlainTextExecutorResult,
)


__all__ = [
    "authorization_header",
    "aiohttp_executor",
    "default_executor",
    "pyodide_executor",
    "urllib_executor",
    "default_executor",
    "Executor",
    "ExecutorError",
    "ExecutorResult",
    "JsonExecutorResult",
    "JsonSeqExecutorResult",
    "PlainTextExecutorResult",
]


def aiohttp_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Returns an `aiohttp`-powered executor"""
    from .aiohttp import AiohttpExecutor

    return AiohttpExecutor(endpoint, authorization)


def pyodide_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Returns a Pyodide-compatible executor"""
    from .pyodide import PyodideExecutor

    return PyodideExecutor(endpoint=endpoint, authorization=authorization)


def urllib_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Returns a native `urllib` executor"""
    from .urllib import UrllibExecutor

    return UrllibExecutor(endpoint=endpoint, authorization=authorization)


def _is_using_pyodide():
    # https://pyodide.org/en/stable/usage/faq.html#how-to-detect-that-code-is-run-with-pyodide
    return "pyodide" in sys.modules


def default_executor(
    endpoint: str, authorization: Optional[str] = None
) -> Executor:
    """Infers the best executor for the current environment"""
    if _is_using_pyodide():
        return pyodide_executor(endpoint=endpoint, authorization=authorization)
    try:
        return aiohttp_executor(endpoint=endpoint, authorization=authorization)
    except ImportError:
        return urllib_executor(endpoint=endpoint, authorization=authorization)


def authorization_header(token: str) -> str:
    """Generates a suitable authorization header from a token"""
    if " " in token:
        return token
    if ":" in token:
        value = base64.b64encode(token.encode("utf8")).decode("utf8")
        return f"Basic {value}"
    return f"Bearer {token}"



from __future__ import annotations

import enum
import json
import logging
import os
from typing import Any, Dict, Optional, cast

from ..common import format_percent, Json, json_dict
from ..data.outcomes import FeasibleOutcome
from ..data.outlines import Label, Outline, outline_from_json
from ..data.solves import SolveInputs
from ..data.tensors import DimensionArgument, Tensor
from ..executors import Executor, JsonExecutorResult
from ..specifications import FormulationSpecification
from ..transformations import Transformation, TransformationContext


DEFAULT_ENDPOINT = "https://api.beta.opvious.io"


class ClientSetting(enum.Enum):
    """Client configuration environment variables"""

    TOKEN = ("OPVIOUS_TOKEN", "")
    ENDPOINT = ("OPVIOUS_ENDPOINT", DEFAULT_ENDPOINT)

    def read(self, env: Optional[dict[str, str]] = None) -> str:
        """Read the setting's current value or default if missing

        Args:
            env: Environment, defaults to `os.environ`.
        """
        if env is None:
            env = cast(Any, os.environ)
        name, default_value = self.value
        return env.get(name) or default_value


def log_progress(logger: logging.Logger, progress: Json) -> None:
    kind = progress["kind"]
    if kind == "activity":
        iter_count = progress.get("lpIterationCount")
        gap = progress.get("relativeGap")
        if iter_count is not None:
            logger.info(
                "Solve in progress... [iterations=%s, gap=%s]",
                iter_count,
                "n/a" if gap is None else format_percent(gap),
            )
    elif kind == "epsilonConstraint":
        logger.info(
            "Added epsilon constraint. [objective_value=%s]",
            progress["objectiveValue"],
        )
    else:
        raise Exception(f"Unsupported progress kind: {kind}")


class SolveInputsBuilder:
    def __init__(self, outline: Outline):
        self._outline = outline
        self._dimensions: Dict[Label, Any] = {}
        self._parameters: Dict[Label, Any] = {}
        self.parameter_entry_count = 0

    def set_dimension(self, label: Label, arg: DimensionArgument) -> None:
        outline = self._outline.dimensions.get(label)
        if not outline:
            raise Exception(f"Unknown dimension: {label}")
        if label in self._dimensions:
            raise Exception(f"Duplicate dimension: {label}")
        items = list(arg)
        self._dimensions[label] = {"label": label, "items": items}

    def set_parameter(self, label: Label, arg: Any) -> None:
        outline = self._outline.parameters.get(label)
        if not outline:
            raise Exception(f"Unknown parameter: {label}")
        if label in self._parameters:
            raise Exception(f"Duplicate parameter: {label}")
        try:
            tensor = Tensor.from_argument(
                arg,
                rank=len(outline.bindings),
                is_indicator=outline.is_indicator,
                is_pin=outline.derivation_kind == "pinnedVariable",
            )
        except Exception as exc:
            raise ValueError(f"Invalid  parameter: {label}") from exc
        self._parameters[label] = json_dict(
            label=label,
            entries=tensor.entries,
            default_value=tensor.default_value,
        )
        self.parameter_entry_count += len(tensor.entries)

    def build(self) -> SolveInputs:
        missing_labels = set()

        for label in self._outline.parameters:
            if label not in self._parameters:
                missing_labels.add(label)

        if self._dimensions:
            for label in self._outline.dimensions:
                if label not in self._dimensions:
                    missing_labels.add(label)

        if missing_labels:
            raise Exception(f"Missing label(s): {missing_labels}")

        return SolveInputs(
            outline=self._outline,
            raw_parameters=list(self._parameters.values()),
            raw_dimensions=list(self._dimensions.values()) or None,
        )


class OutlineGenerator:
    def __init__(self, executor: Executor, outline_data: Json):
        self._executor = executor
        self._pristine_outline_data = outline_data
        self._transformations = cast(list[Transformation], [])

    @classmethod
    async def formulation(
        cls, executor: Executor, specification: FormulationSpecification
    ) -> tuple[OutlineGenerator, str]:
        data = await executor.execute_graphql_query(
            query="@FetchOutline",
            variables={
                "formulationName": specification.formulation_name,
                "tagName": specification.tag_name,
            },
        )
        formulation = data.get("formulation")
        if not formulation:
            raise Exception("No matching formulation found")
        tag = formulation.get("tag")
        if not tag:
            raise Exception("No matching specification found")
        spec = tag["specification"]
        return (OutlineGenerator(executor, spec["outline"]), tag["name"])

    @classmethod
    async def sources(
        cls, executor: Executor, sources: list[str]
    ) -> OutlineGenerator:
        async with executor.execute(
            result_type=JsonExecutorResult,
            url="/sources/parse",
            method="POST",
            json_data=json_dict(sources=sources, outline=True),
        ) as res:
            outline_data = res.json_data()
        errors = outline_data.get("errors")
        if errors:
            raise Exception(f"Invalid sources: {json.dumps(errors)}")
        return OutlineGenerator(executor, outline_data["outline"])

    def add_transformation(self, tf: Transformation) -> None:
        self._transformations.append(tf)

    async def generate(self) -> tuple[Outline, Json]:
        pristine_outline = outline_from_json(self._pristine_outline_data)
        if not self._transformations:
            return pristine_outline, []
        executor = self._executor
        pristine_outline_data = self._pristine_outline_data

        class Context(TransformationContext):
            async def fetch_outline(self) -> Outline:
                transformations = self.get_json()
                if not transformations:
                    return pristine_outline
                async with executor.execute(
                    result_type=JsonExecutorResult,
                    url="/outlines/transform",
                    method="POST",
                    json_data=json_dict(
                        outline=pristine_outline_data,
                        transformations=transformations,
                    ),
                ) as res:
                    data = res.json_data()
                return outline_from_json(data["outline"])

        context = Context()
        for tf in self._transformations:
            await tf.register(context)
        outline = await context.fetch_outline()
        return outline, context.get_json()


def feasible_outcome_details(outcome: FeasibleOutcome) -> Optional[str]:
    details = []
    if outcome.objective_value:
        details.append(f"objective={outcome.objective_value}")
    if outcome.relative_gap:
        details.append(f"gap={format_percent(outcome.relative_gap)}")
    return ", ".join(details) if details else None
